==> sage-v0.3-2005-05-02-src.tar: sage-0.3/sage/tables/elliptic_curves.py <==
"""
Table of elliptic curves of conductor <= 30000.
"""

#*****************************************************************************
#       SAGE: System for Arithmetic Geometry Experimentation
#
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import os, weakref

import sage.tables.db   # very important that this be fully qualified
import sage.misc.errors as errors
import sage.misc.misc as misc
from   sage.rings.arith import factor
from sage.pari.py_pari import eval as p_eval
from sage.rings.rings import Integer

class EllipticCurve:
    """
    Class that defines an elliptic curve in the database.
    """
    def __init__(self, conductor, iso_class, number,
                 a_invariants,
                 rank=None,
                 torsion=None,
                 tamagawa=None,
                 modular_degree=None,
                 omega=None,
                 lseries1=None,
                 regulator=None,
                 sha_an=None):
        self.conductor = conductor
        self.iso_class = iso_class
        self.number = number
        self.a_invariants = a_invariants
        self.rank = rank
        self.torsion = torsion
        self.tamagawa = tamagawa   # {p: (p, cp), ...}
        self.modular_degree = modular_degree
        self.omega = omega
        self.lseries1 = lseries1
        self.regulator = regulator
        self.sha_an = sha_an

    def __repr__(self):
        s = "%s%s%s: invs=%s, rank=%s, tors=%s, cp=%s"%(\
            self.conductor, self.iso_class, self.number, self.a_invariants,
            self.rank, self.torsion, self.tamagawa)
        try:
            s += ", degphi=%s"%self.modular_degree
        except AttributeError:
            pass
        return s

    def tamagawa_product(self):
        x = 1
        for cp in self.tamagawa.itervalues():
            x *= cp
        return x

    def traces(self,n):
        cmd = ("ellan(ellinit(%s),%s)"%(self.a_invariants,n)).replace("L","")
        return eval(p_eval(cmd))
    anlist = traces

    

class Database(sage.tables.db.Database):
    def __init__(self, read_only=True):
        """
        INPUT:
            read_only -- bool, default: True, if False then
                         can change the database; otherwise, not.
        """
        sage.tables.db.Database.__init__(self,
             name="elliptic_curves", read_only=read_only)

    def __repr__(self):
        return "Extended table of elliptic curves based on Cremona's data."

    def __getslice__(self, i, j):
        #j = min(j, max(self.keys())+1)
        return [self[k]  for k in xrange(i,j)]

    def __getitem__(self, i):
        if i <= 10:
            return []
        try:
            return sage.tables.db.Database.__getitem__(self, i)
        except KeyError:
            return []

    def __iter__(self):
        """
        Iterate through all optimal curves in the database.
        """
        N = 11
        while True:
            for C in self[N]:
                yield C[0]
            N += 1
        
    def curves(self, conductors):
        """
        Return the list of all curves in the database with given
        conductors.

        INPUT:
            conductor - an iterable
        """
        if isinstance(conductors, (int, Integer)):
            conductors = [conductors]
        return sum(sum([self[x] for x in conductors], []), [])
    all_curves = curves

    def optimal_curves(self, conductors):
        """
        Return the list of all optimal curves in the database
        with given conductor(s).

        INPUT:
            conductor - an integer.
        """
        if isinstance(conductors, (int,Integer)):
            conductors = [conductors]
        return sum([[X[0] for X in self[N]] for N in conductors], [])


    def curve(self, label):
        """
        Return the curve with given Cremona label.
        """
        if not isinstance(label, str):
            label = str(label)
        i=0
        while i<len(label) and label[i]>='0' and label[i]<='9':
            i += 1
        j=i+1
        if j>len(label):
            label += "A"
        while j<len(label) and not (label[j]>='0' and label[j]<='9'):
            j += 1
        if j>=len(label):
            label += "1"
        level, iso, num = int(label[:i]), label[i:j], int(label[j:])
        iso = iso.upper()
        iso = (ord(iso[0]) - ord('A')) + 26*(len(iso)-1)
        try:
            return self[level][iso][num-1]
        except (KeyError, IndexError):
            raise ArithmeticError, "There is no elliptic curve with Cremona label '%s'."%label

    def _import_allcurves(self, file, start, stop):
        """
        Load data from one of Cremona's allcurves files, from conductor start to stop.
        """
        #import sage.tables.elliptic_curves
        level = []
        clss = []
        N_last = 0
        for line in open(file).readlines():
            X = line.split()
            if len(X) < 6:
                continue
            N, iso, num, ainvs, rank, tor = line.split()
            N = int(N)
            if N < start:
                continue
            print line
            if N != N_last and N > 0:
                print "Saving level", N_last
                self[N_last] = level
                level = []; clss = []
                level.append(clss)
                N_last = N
            if N > stop:
                return
            if N_last == 0:
                N_last = N
            num = int(num)
            ainvs = eval(ainvs)
            rank = int(rank)
            tor = int(tor)
            P = [p for p, _ in factor(N)]
            cp = {}
            for p in P:
                cmd = "elllocalred(ellinit(%s),%s)"%(ainvs, p)
                cmd = cmd.replace("L","")
                x = eval(p_eval(cmd))
                cp[p] = x[3]
            E = sage.tables.elliptic_curves.EllipticCurve(conductor = N,
                             iso_class = iso,
                             number = num,
                             a_invariants = ainvs,
                             rank = rank,
                             torsion = tor,
                             tamagawa = cp)

            if num == 1 and len(clss) > 0:
                clss = []
                level.append(clss)
            clss.append(E)
            print "clss = ", clss
        self[N] = level
        self.commit()

    def _import_degphi(self, file, start, stop):
        """
        Load data from one of Cremona's degphi files, from
        conductor start to stop.
        """
        def emit(conductor, data):
            print "Level", conductor
            C = self[conductor]
            assert len(C) == len(data)
            for i in xrange(len(C)):
                E = C[i][0]
                if conductor != 990:   # known problem for 990H
                    assert E.a_invariants == data[i][3]
                assert E.iso_class == data[i][1]
                assert E.conductor == data[i][0]
                E.modular_degree = data[i][2]

        data = []
        N_last = 0
        first = True
        for line in open(file).readlines():
            print line
            X = line.split()
            if len(X) < 5:
                continue
            N, iso, degphi, _, model = line.split()
            assert iso[-1] == "1"
            N = int(N)
            if N < start:
                continue
            if N != N_last and N_last > 0:
                emit(N_last, data)
                data = []
                N_last = N
            if N > stop:
                return
            if N_last == 0:
                N_last = N
            iso = iso[:-1]
            degphi = int(degphi)
            model = eval(model)
            data.append([N, iso, degphi, model])
        emit(N, data)
        self.commit()

    def _import_allbsd(self, file, start, stop):
        """
        Load data from one of Cremona's allbsd files, from
        conductor start to stop.
        """
        def emit(conductor, data):
            print "Level", conductor
            C = self.curves(conductor)
            for i, E in enumerate(C):
                F = data[i]
                assert E.conductor == F[0]
                assert E.iso_class == F[1]
                assert E.number == F[2]
                if conductor != 990:   # known problem for 990H
                    assert E.a_invariants == F[3]
                E.omega     = F[4]
                E.lseries1  = F[5]
                E.regulator = F[6]
                E.sha_an    = F[7]
            self[conductor] = self[conductor]   # tell database that this entry changed.

        data = []
        N_last = 0
        first = True
        for line in open(file).readlines():
            print line
            X = line.split()
            if len(X) < 11:
                continue
            t = tuple(list(line.split())[:11])
            N, iso, num, ainvs, rank, tors, cp, om, L, Reg, Sha = t
            N = int(N)
            if N < start:
                continue
            if N != N_last and N_last > 0:
                emit(N_last, data)
                data = []
                N_last = N
            if N > stop:
                return
            if N_last == 0:
                N_last = N
            data.append([N, iso, int(num), eval(ainvs), eval(om), eval(L), eval(Reg), eval(Sha)])
        emit(N, data)

        self.commit()

==> sage-v0.3-2005-05-09-src.tar: sage-0.3/sage/tables/elliptic_curves.py <==
"""
Table of elliptic curves of conductor <= 30000.
"""

#*****************************************************************************
#       SAGE: System for Arithmetic Geometry Experimentation
#
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import os, weakref

import sage.tables.db   # very important that this be fully qualified
import sage.misc.errors as errors
import sage.misc.misc as misc
from   sage.rings.arith import factor
from sage.pari.py_pari import eval as p_eval
from sage.rings.rings import Integer

class EllipticCurve:
    """
    Class that defines an elliptic curve in the database.
    """
    def __init__(self, conductor, iso_class, number,
                 a_invariants,
                 rank=None,
                 torsion=None,
                 tamagawa=None,
                 modular_degree=None,
                 omega=None,
                 lseries1=None,
                 regulator=None,
                 sha_an=None):
        self.conductor = conductor
        self.iso_class = iso_class
        self.number = number
        self.a_invariants = a_invariants
        self.rank = rank
        self.torsion = torsion
        self.tamagawa = tamagawa   # {p: (p, cp), ...}
        self.modular_degree = modular_degree
        self.omega = omega
        self.lseries1 = lseries1
        self.regulator = regulator
        self.sha_an = sha_an

    def __repr__(self):
        s = "%s%s%s: invs=%s, rank=%s, tors=%s, cp=%s"%(\
            self.conductor, self.iso_class, self.number, self.a_invariants,
            self.rank, self.torsion, self.tamagawa)
        try:
            s += ", degphi=%s"%self.modular_degree
        except AttributeError:
            pass
        try:
            s += ", reg=%s"%self.regulator
        except AttributeError:
            pass
        return s

    def tamagawa_product(self):
        x = 1
        for cp in self.tamagawa.itervalues():
            x *= cp
        return x

    def traces(self,n):
        cmd = ("ellan(ellinit(%s),%s)"%(self.a_invariants,n)).replace("L","")
        return eval(p_eval(cmd))
    anlist = traces

    

class Database(sage.tables.db.Database):
    """
    INPUT:
        read_only -- bool, default: True, if False then
                     can change the database; otherwise, not.
    """
    def __init__(self, read_only=True):
        sage.tables.db.Database.__init__(self,
             name="elliptic_curves", read_only=read_only)

    def __repr__(self):
        return "Extended table of elliptic curves based on Cremona's data."

    def __getslice__(self, i, j):
        #j = min(j, max(self.keys())+1)
        return [self[k]  for k in xrange(i,j)]

    def __getitem__(self, i):
        if i <= 10:
            return []
        try:
            return sage.tables.db.Database.__getitem__(self, i)
        except KeyError:
            return []

    def __iter__(self):
        """
        Iterate through all optimal curves in the database.
        """
        N = 11
        while True:
            for C in self[N]:
                yield C[0]
            N += 1
        
    def curves(self, conductors):
        """
        Return the list of all curves in the database with given
        conductors.

        INPUT:
            conductor - an iterable
        """
        if isinstance(conductors, (int, Integer)):
            conductors = [conductors]
        return sum(sum([self[x] for x in conductors], []), [])
    all_curves = curves

    def optimal_curves(self, conductors):
        """
        Return the list of all optimal curves in the database
        with given conductor(s).

        INPUT:
            conductor - an integer.
        """
        if isinstance(conductors, (int,Integer)):
            conductors = [conductors]
        return sum([[X[0] for X in self[N]] for N in conductors], [])


    def curve(self, label):
        """
        Return the curve with given Cremona label.
        """
        if not isinstance(label, str):
            label = str(label)
        i=0
        while i<len(label) and label[i]>='0' and label[i]<='9':
            i += 1
        j=i+1
        if j>len(label):
            label += "A"
        while j<len(label) and not (label[j]>='0' and label[j]<='9'):
            j += 1
        if j>=len(label):
            label += "1"
        level, iso, num = int(label[:i]), label[i:j], int(label[j:])
        iso = iso.upper()
        iso = (ord(iso[0]) - ord('A')) + 26*(len(iso)-1)
        try:
            return self[level][iso][num-1]
        except (KeyError, IndexError):
            raise ValueError, "There is no elliptic curve with Cremona label '%s'."%label

    def isogeny_class(self, label):
        """
        Return the isogeny class of the curve with given Cremona label.
        """
        if not isinstance(label, str):
            label = str(label)
        i=0
        while i<len(label) and label[i]>='0' and label[i]<='9':
            i += 1
        j=i+1
        if j>len(label):
            label += "A"
        while j<len(label) and not (label[j]>='0' and label[j]<='9'):
            j += 1
        if j>=len(label):
            label += "1"
        level, iso, _ = int(label[:i]), label[i:j], int(label[j:])
        iso = iso.upper()
        iso = (ord(iso[0]) - ord('A')) + 26*(len(iso)-1)
        try:
            return self[level][iso]
        except (KeyError, IndexError):
            raise ValueError, "There is no elliptic curve with Cremona label '%s'."%label

    def _import_allcurves(self, file, start, stop):
        """
        Load data from one of Cremona's allcurves files, from conductor start to stop.
        """
        #import sage.tables.elliptic_curves
        level = []
        clss = []
        N_last = 0
        for line in open(file).readlines():
            X = line.split()
            if len(X) < 6:
                continue
            N, iso, num, ainvs, rank, tor = line.split()
            N = int(N)
            if N < start:
                continue
            print line
            if N != N_last and N > 0:
                print "Saving level", N_last
                self[N_last] = level
                level = []; clss = []
                level.append(clss)
                N_last = N
            if N > stop:
                return
            if N_last == 0:
                N_last = N
            num = int(num)
            ainvs = eval(ainvs)
            rank = int(rank)
            tor = int(tor)
            P = [p for p, _ in factor(N)]
            cp = {}
            for p in P:
                cmd = "elllocalred(ellinit(%s),%s)"%(ainvs, p)
                cmd = cmd.replace("L","")
                x = eval(p_eval(cmd))
                cp[p] = x[3]
            E = sage.tables.elliptic_curves.EllipticCurve(conductor = N,
                             iso_class = iso,
                             number = num,
                             a_invariants = ainvs,
                             rank = rank,
                             torsion = tor,
                             tamagawa = cp)

            if num == 1 and len(clss) > 0:
                clss = []
                level.append(clss)
            clss.append(E)
            print "clss = ", clss
        self[N] = level
        self.commit()

    def _import_degphi(self, file, start, stop):
        """
        Load data from one of Cremona's degphi files, from
        conductor start to stop.
        """
        def emit(conductor, data):
            print "Level", conductor
            C = self[conductor]
            assert len(C) == len(data)
            for i in xrange(len(C)):
                E = C[i][0]
                if conductor != 990:   # known problem for 990H
                    assert E.a_invariants == data[i][3]
                assert E.iso_class == data[i][1]
                assert E.conductor == data[i][0]
                E.modular_degree = data[i][2]

        data = []
        N_last = 0
        first = True
        for line in open(file).readlines():
            print line
            X = line.split()
            if len(X) < 5:
                continue
            N, iso, degphi, _, model = line.split()
            assert iso[-1] == "1"
            N = int(N)
            if N < start:
                continue
            if N != N_last and N_last > 0:
                emit(N_last, data)
                data = []
                N_last = N
            if N > stop:
                return
            if N_last == 0:
                N_last = N
            iso = iso[:-1]
            degphi = int(degphi)
            model = eval(model)
            data.append([N, iso, degphi, model])
        emit(N, data)
        self.commit()

    def _import_allbsd(self, file, start, stop):
        """
        Load data from one of Cremona's allbsd files, from
        conductor start to stop.
        """
        def emit(conductor, data):
            print "Level", conductor
            C = self.curves(conductor)
            for i, E in enumerate(C):
                F = data[i]
                assert E.conductor == F[0]
                assert E.iso_class == F[1]
                assert E.number == F[2]
                if conductor != 990:   # known problem for 990H
                    assert E.a_invariants == F[3]
                E.omega     = F[4]
                E.lseries1  = F[5]
                E.regulator = F[6]
                E.sha_an    = F[7]
            self[conductor] = self[conductor]   # tell database that this entry changed.

        data = []
        N_last = 0
        first = True
        for line in open(file).readlines():
            print line
            X = line.split()
            if len(X) < 11:
                continue
            t = tuple(list(line.split())[:11])
            N, iso, num, ainvs, rank, tors, cp, om, L, Reg, Sha = t
            N = int(N)
            if N < start:
                continue
            if N != N_last and N_last > 0:
                emit(N_last, data)
                data = []
                N_last = N
            if N > stop:
                return
            if N_last == 0:
                N_last = N
            data.append([N, iso, int(num), eval(ainvs), eval(om), eval(L), eval(Reg), eval(Sha)])
        emit(N, data)

        self.commit()
